@@ -14,13 +14,12 @@ import time
 from types import FrameType
 from typing import Any, Callable, List, Optional, Tuple, TypeVar, Union
 
-import _frida
-
 if platform.system() == "Windows":
     import msvcrt
 
 import colorama
 import frida
+import frida._frida as _frida
 
 from frida_tools.reactor import Reactor
 
@@ -106,9 +105,9 @@ def create_target_parser(target_type: str) -> Callable[[str], TargetTypeTuple]:
     def parse_target(value: str) -> TargetTypeTuple:
         if target_type == "file":
             return (target_type, [value])
-        elif target_type == "gated":
+        if target_type == "gated":
             return (target_type, re.compile(value))
-        elif target_type == "pid":
+        if target_type == "pid":
             return (target_type, int(value))
         return (target_type, value)
 
@@ -147,7 +146,10 @@ class ConsoleApplication:
         if hasattr(signal, "SIGPIPE"):
             signal.signal(signal.SIGPIPE, signal.SIG_DFL)
 
-        colorama.init(strip=True if plain_terminal else None)
+        # If true, emit text without colors.  https://no-color.org/
+        no_color = plain_terminal or bool(os.environ.get("NO_COLOR"))
+
+        colorama.init(strip=True if no_color else None)
 
         parser = self._initialize_arguments_parser()
         real_args = compute_real_args(parser, args=args)
@@ -420,6 +422,19 @@ class ConsoleApplication:
         frida.shutdown()
         sys.exit(self._exit_status)
 
+    def _respawn(self) -> None:
+        self._session.off("detached", self._schedule_on_session_detached)
+        self._stop()
+        self._session = None
+
+        self._device.kill(self._spawned_pid)
+        self._spawned_pid = None
+        self._spawned_argv = None
+        self._resumed = False
+
+        self._attach_and_instrument()
+        self._resume()
+
     def _add_options(self, parser: argparse.ArgumentParser) -> None:
         """
         override this method if you want to add custom arguments to your
@@ -523,6 +538,9 @@ class ConsoleApplication:
         self._on_device_found()
         self._device.on("output", self._schedule_on_output)
         self._device.on("lost", self._schedule_on_device_lost)
+        self._attach_and_instrument()
+
+    def _attach_and_instrument(self) -> None:
         if self._target is not None:
             target_type, target_value = self._target
 
@@ -599,6 +617,15 @@ class ConsoleApplication:
         self._start()
         self._started = True
 
+    def _pick_worker_pid(self) -> int:
+        try:
+            frontmost = self._device.get_frontmost_application()
+            if frontmost is not None and frontmost.identifier == "re.frida.Gadget":
+                return frontmost.pid
+        except:
+            pass
+        return 0
+
     def _attach(self, pid: int) -> None:
         self._target_pid = pid
 
