@@ -271,6 +271,8 @@ class REPLApplication(ConsoleApplication):
             self._logfile.close()
 
         if self._kill_on_exit and self._spawned_pid is not None:
+            if self._session is not None:
+                self._session.detach()
             self._device.kill(self._spawned_pid)
 
         if not self._quiet:
@@ -301,11 +303,11 @@ class REPLApplication(ConsoleApplication):
             if isinstance(cmodule_code, bytes):
                 script.post({"type": "frida:cmodule-payload"}, data=cmodule_code)
                 cmodule_code = None
-            script.exports.frida_load_cmodule(cmodule_code, self._toolchain)
+            script.exports_sync.frida_load_cmodule(cmodule_code, self._toolchain)
 
         stage = "early" if self._target[0] == "file" and is_first_load else "late"
         try:
-            script.exports.init(stage, self._user_parameters)
+            script.exports_sync.init(stage, self._user_parameters)
         except:
             pass
 
@@ -486,8 +488,9 @@ class REPLApplication(ConsoleApplication):
 
             stack = error.get("stack", None)
             if stack is not None:
-                trim_amount = 5 if self._runtime == "v8" else 6
-                trimmed_stack = stack.split("\n")[1:-trim_amount]
+                message_len = len(error["message"].split("\n"))
+                trim_amount = 6 if self._runtime == "v8" else 7
+                trimmed_stack = stack.split("\n")[message_len:-trim_amount]
                 if len(trimmed_stack) > 0:
                     output += "\n" + "\n".join(trimmed_stack)
         except frida.InvalidOperationError:
@@ -616,14 +619,18 @@ class REPLApplication(ConsoleApplication):
 
     def _is_java_available(self) -> bool:
         assert self._session is not None
-        script = self._session.create_script(
-            name="java_check", source="rpc.exports.javaAvailable = () => Java.available;", runtime=self._runtime
-        )
-        script.load()
+        script = None
         try:
-            return script.exports.java_available()
+            script = self._session.create_script(
+                name="java_check", source="rpc.exports.javaAvailable = () => Java.available;", runtime=self._runtime
+            )
+            script.load()
+            return script.exports_sync.java_available()
         except:
             return False
+        finally:
+            if script is not None:
+                script.unload()
 
     def _refresh_prompt(self) -> None:
         self._prompt_string = self._create_prompt()
@@ -655,12 +662,12 @@ class REPLApplication(ConsoleApplication):
 
     def _evaluate_expression(self, expression: str) -> Tuple[str, bytes]:
         assert self._script is not None
-        result = self._script.exports.frida_evaluate_expression(expression)
+        result = self._script.exports_sync.frida_evaluate_expression(expression)
         return self._parse_evaluate_result(result)
 
     def _evaluate_quick_command(self, tokens: List[str]) -> Tuple[str, bytes]:
         assert self._script is not None
-        result = self._script.exports.frida_evaluate_quick_command(tokens)
+        result = self._script.exports_sync.frida_evaluate_quick_command(tokens)
         return self._parse_evaluate_result(result)
 
     def _parse_evaluate_result(self, result: Union[bytes, Mapping[Any, Any], Tuple[str, bytes]]) -> Tuple[str, bytes]:
@@ -838,19 +845,19 @@ try {
 }
 
 Object.defineProperty(rpc, 'exports', {
-get() {
-    return rpcExports;
-},
-set(value) {
-    for (const [k, v] of Object.entries(value)) {
-        rpcExports[k] = v;
+    get() {
+        return rpcExports;
+    },
+    set(value) {
+        for (const [k, v] of Object.entries(value)) {
+            rpcExports[k] = v;
+        }
     }
-}
 });
 
 function onLog(messagePtr) {
-const message = messagePtr.readUtf8String();
-console.log(message);
+    const message = messagePtr.readUtf8String();
+    console.log(message);
 }
 """
 
@@ -877,18 +884,18 @@ extern void _frida_log (const gchar * message);
 
 static void
 frida_log (const char * format,
-       ...)
+           ...)
 {
-gchar * message;
-va_list args;
+  gchar * message;
+  va_list args;
 
-va_start (args, format);
-message = g_strdup_vprintf (format, args);
-va_end (args);
+  va_start (args, format);
+  message = g_strdup_vprintf (format, args);
+  va_end (args);
 
-_frida_log (message);
+  _frida_log (message);
 
-g_free (message);
+  g_free (message);
 }
 """
         )
