@@ -1,19 +1,40 @@
+from __future__ import annotations
+
 import argparse
+import asyncio
 import binascii
 import codecs
+import email.utils
+import errno
+import gzip
+import http
+import mimetypes
 import os
-import platform
 import re
+import shlex
 import subprocess
-from typing import Callable, List, Optional, Union
+import threading
+from collections import OrderedDict
+from dataclasses import dataclass
+from pathlib import Path
+from typing import Callable, Generator, List, Optional, Set
+from zipfile import ZipFile
 
 import frida
+import websockets.asyncio.server
+import websockets.datastructures
+import websockets.exceptions
+import websockets.http11
 
 from frida_tools.reactor import Reactor
 
+MANPAGE_CONTROL_CHARS = re.compile(r"\.[a-zA-Z]*(\s|$)|\s?\"")
+MANPAGE_FUNCTION_PROTOTYPE = re.compile(r"([a-zA-Z_]\w+)\(([^\)]+)")
+
 
 def main() -> None:
     import json
+    import traceback
 
     from colorama import Fore, Style
 
@@ -22,9 +43,15 @@ def main() -> None:
     class TracerApplication(ConsoleApplication, UI):
         def __init__(self) -> None:
             super().__init__(await_ctrl_c)
-            self._palette = [Fore.CYAN, Fore.MAGENTA, Fore.YELLOW, Fore.GREEN, Fore.RED, Fore.BLUE]
+            self._handlers = OrderedDict()
+            self._ui_port = 1337
+            self._ui_zip = ZipFile(Path(__file__).parent / "tracer_ui.zip", "r")
+            self._ui_socket_handlers: Set[UISocketHandler] = set()
+            self._ui_worker = None
+            self._asyncio_loop = None
+            self._palette = ["cyan", "magenta", "yellow", "green", "red", "blue"]
             self._next_color = 0
-            self._attributes_by_thread_id = {}
+            self._style_by_thread_id = {}
             self._last_event_tid = -1
 
         def _add_options(self, parser: argparse.ArgumentParser) -> None:
@@ -66,6 +93,20 @@ def main() -> None:
                 metavar="OBJC_METHOD",
                 type=pb.exclude_objc_method,
             )
+            parser.add_argument(
+                "-y",
+                "--include-swift-func",
+                help="include SWIFT_FUNC",
+                metavar="SWIFT_FUNC",
+                type=pb.include_swift_func,
+            )
+            parser.add_argument(
+                "-Y",
+                "--exclude-swift-func",
+                help="exclude SWIFT_FUNC",
+                metavar="SWIFT_FUNC",
+                type=pb.exclude_swift_func,
+            )
             parser.add_argument(
                 "-j",
                 "--include-java-method",
@@ -118,6 +159,7 @@ def main() -> None:
             return "%(prog)s [options] target"
 
         def _initialize(self, parser: argparse.ArgumentParser, options: argparse.Namespace, args: List[str]) -> None:
+            self._repo: Optional[FileRepository] = None
             self._tracer: Optional[Tracer] = None
             self._profile = self._profile_builder.build()
             self._quiet: bool = options.quiet
@@ -146,14 +188,20 @@ def main() -> None:
             return True
 
         def _start(self) -> None:
+            if self._ui_worker is None:
+                worker = threading.Thread(target=self._run_ui_server, name="ui-server", daemon=True)
+                worker.start()
+                self._ui_worker = worker
+
             if self._output_path is not None:
                 self._output = OutputFile(self._output_path)
 
             stage = "early" if self._target[0] == "file" else "late"
 
+            self._repo = FileRepository(self._reactor, self._decorate)
             self._tracer = Tracer(
                 self._reactor,
-                FileRepository(self._reactor, self._decorate),
+                self._repo,
                 self._profile,
                 self._init_scripts,
                 log_handler=self._log,
@@ -163,6 +211,7 @@ def main() -> None:
             except Exception as e:
                 self._update_status(f"Failed to start tracing: {e}")
                 self._exit(1)
+                return
 
         def _stop(self) -> None:
             self._tracer.stop()
@@ -171,6 +220,11 @@ def main() -> None:
                 self._output.close()
             self._output = None
 
+            self._handlers.clear()
+            self._next_color = 0
+            self._style_by_thread_id.clear()
+            self._last_event_tid = -1
+
         def on_script_created(self, script: frida.core.Script) -> None:
             self._on_script_created(script)
 
@@ -185,7 +239,9 @@ def main() -> None:
                     plural = ""
                 else:
                     plural = "s"
-                self._update_status("Started tracing %d function%s. Press Ctrl+C to stop." % (count, plural))
+                self._update_status(
+                    f"Started tracing {count} function{plural}. Web UI available at http://localhost:{self._ui_port}/"
+                )
 
         def on_trace_warning(self, message: str) -> None:
             self._print(Fore.RED + Style.BRIGHT + "Warning" + Style.RESET_ALL + ": " + message)
@@ -194,41 +250,232 @@ def main() -> None:
             self._print(Fore.RED + Style.BRIGHT + "Error" + Style.RESET_ALL + ": " + message)
             self._exit(1)
 
-        def on_trace_events(self, events) -> None:
+        def on_trace_events(self, raw_events) -> None:
+            events = [
+                (target_id, timestamp, thread_id, depth, caller, backtrace, message, self._get_style(thread_id))
+                for target_id, timestamp, thread_id, depth, caller, backtrace, message in raw_events
+            ]
+            self._asyncio_loop.call_soon_threadsafe(
+                lambda: self._asyncio_loop.create_task(self._broadcast_trace_events(events))
+            )
+
             no_attributes = Style.RESET_ALL
-            for timestamp, thread_id, depth, message in events:
+            for target_id, timestamp, thread_id, depth, caller, backtrace, message, style in events:
                 if self._output is not None:
                     self._output.append(message + "\n")
                 elif self._quiet:
                     self._print(message)
                 else:
                     indent = depth * "   | "
-                    attributes = self._get_attributes(thread_id)
+                    attributes = getattr(Fore, style[0].upper())
+                    if len(style) > 1:
+                        attributes += getattr(Style, style[1].upper())
                     if thread_id != self._last_event_tid:
                         self._print("%s           /* TID 0x%x */%s" % (attributes, thread_id, Style.RESET_ALL))
                         self._last_event_tid = thread_id
                     self._print("%6d ms  %s%s%s%s" % (timestamp, attributes, indent, message, no_attributes))
 
-        def on_trace_handler_create(self, target, handler, source) -> None:
+        def on_trace_handler_create(self, target: TraceTarget, handler: str, source: str) -> None:
+            self._register_handler(target, source)
             if self._quiet:
                 return
             self._print('%s: Auto-generated handler at "%s"' % (target, source.replace("\\", "\\\\")))
 
-        def on_trace_handler_load(self, target, handler, source) -> None:
+        def on_trace_handler_load(self, target: TraceTarget, handler: str, source: str) -> None:
+            self._register_handler(target, source)
             if self._quiet:
                 return
             self._print('%s: Loaded handler at "%s"' % (target, source.replace("\\", "\\\\")))
 
-        def _get_attributes(self, thread_id):
-            attributes = self._attributes_by_thread_id.get(thread_id, None)
-            if attributes is None:
+        def _register_handler(self, target: TraceTarget, source: str) -> None:
+            config = {"capture_backtraces": False}
+            self._handlers[target.identifier] = (target, source, config)
+
+        def _get_style(self, thread_id):
+            style = self._style_by_thread_id.get(thread_id, None)
+            if style is None:
                 color = self._next_color
                 self._next_color += 1
-                attributes = self._palette[color % len(self._palette)]
+                style = [self._palette[color % len(self._palette)]]
                 if (1 + int(color / len(self._palette))) % 2 == 0:
-                    attributes += Style.BRIGHT
-                self._attributes_by_thread_id[thread_id] = attributes
-            return attributes
+                    style.append("bright")
+                self._style_by_thread_id[thread_id] = style
+            return style
+
+        def _run_ui_server(self):
+            asyncio.run(self._handle_ui_requests())
+
+        async def _handle_ui_requests(self):
+            self._asyncio_loop = asyncio.get_running_loop()
+            while True:
+                try:
+                    async with websockets.asyncio.server.serve(
+                        self._handle_websocket_connection,
+                        "localhost",
+                        self._ui_port,
+                        process_request=self._handle_asset_request,
+                    ):
+                        await asyncio.get_running_loop().create_future()
+                        return
+                except OSError as e:
+                    if e.errno == errno.EADDRINUSE:
+                        self._ui_port += 1
+                    else:
+                        raise
+
+        async def _handle_websocket_connection(self, websocket: websockets.asyncio.server.ServerConnection):
+            if self._tracer is None:
+                return
+
+            handler = UISocketHandler(self, websocket)
+            self._ui_socket_handlers.add(handler)
+            try:
+                await handler.process_messages()
+            except:
+                traceback.print_exc()
+                # pass
+            finally:
+                self._ui_socket_handlers.remove(handler)
+
+        async def _broadcast_trace_events(self, events):
+            for handler in self._ui_socket_handlers:
+                await handler.post(
+                    {
+                        "type": "events:add",
+                        "events": events,
+                    }
+                )
+
+        def _handle_asset_request(
+            self, connection: websockets.asyncio.server.ServerConnection, request: websockets.asyncio.server.Request
+        ):
+            if request.headers.get("Connection") == "Upgrade":
+                return
+
+            raw_path = request.path.split("?", maxsplit=1)[0]
+
+            filename = raw_path[1:]
+            if filename == "":
+                filename = "index.html"
+
+            try:
+                body = self._ui_zip.read(filename)
+            except KeyError:
+                return connection.respond(http.HTTPStatus.NOT_FOUND, "File not found\n")
+
+            status = http.HTTPStatus(http.HTTPStatus.OK)
+
+            content_type, content_encoding = mimetypes.guess_type(filename)
+            if content_type is None:
+                content_type = "application/octet-stream"
+
+            headers = websockets.datastructures.Headers(
+                [
+                    ("Connection", "close"),
+                    ("Content-Length", str(len(body))),
+                    ("Content-Type", content_type),
+                    ("Date", email.utils.formatdate(usegmt=True)),
+                ]
+            )
+            if content_encoding is not None:
+                headers.update({"Content-Encoding": content_encoding})
+
+            response = websockets.http11.Response(status.value, status.phrase, headers, body)
+            connection.protocol.handshake_exc = websockets.exceptions.InvalidStatus(response)
+
+            return response
+
+    class UISocketHandler:
+        def __init__(self, app: TracerApplication, socket: websockets.asyncio.server.ServerConnection) -> None:
+            self.app = app
+            self.socket = socket
+
+        async def process_messages(self) -> None:
+            app = self.app
+
+            await self.post(
+                {
+                    "type": "tracer:sync",
+                    "spawned_program": app._spawned_argv[0] if app._spawned_argv is not None else None,
+                    "process": app._tracer.process,
+                    "handlers": [target.to_json() for target, _, _ in app._handlers.values()],
+                }
+            )
+
+            while True:
+                request = json.loads(await self.socket.recv())
+                request_id = request.get("id")
+
+                try:
+                    handle_request = getattr(self, "_on_" + request["type"].replace(":", "_").replace("-", "_"), None)
+                    if handle_request is None:
+                        raise NameError("unsupported request type")
+                    result = await handle_request(request["payload"])
+                except Exception as e:
+                    if request_id is not None:
+                        await self.post(
+                            {
+                                "type": "request:error",
+                                "id": request_id,
+                                "payload": {
+                                    "message": str(e),
+                                    "stack": traceback.format_exc(),
+                                },
+                            }
+                        )
+                    continue
+
+                if request_id is not None:
+                    await self.post({"type": "request:result", "id": request_id, "payload": result})
+
+        async def post(self, message: dict) -> None:
+            await self.socket.send(json.dumps(message))
+
+        async def _on_tracer_respawn(self, _: dict) -> None:
+            self.app._reactor.schedule(self.app._respawn)
+
+        async def _on_handler_load(self, payload: dict) -> None:
+            target, source, config = self.app._handlers[payload["id"]]
+            return {"code": self.app._repo.ensure_handler(target), "config": config}
+
+        async def _on_handler_save(self, payload: dict) -> None:
+            target, _, _ = self.app._handlers[payload["id"]]
+            self.app._repo.update_handler(target, payload["code"])
+
+        async def _on_handler_configure(self, payload: dict) -> None:
+            identifier = payload["id"]
+            _, _, config = self.app._handlers[identifier]
+            for k, v in payload["parameters"].items():
+                config[k] = v
+            self.app._tracer.update_handler_config(identifier, config)
+
+        async def _on_targets_stage(self, payload: dict) -> None:
+            profile = TracerProfile(list(map(tuple, payload["profile"]["spec"])))
+            items = self.app._tracer.stage_targets(profile)
+            return {
+                "items": items,
+            }
+
+        async def _on_targets_commit(self, payload: dict) -> None:
+            result = self.app._tracer.commit_targets(payload["id"])
+            target_ids = result["ids"]
+
+            await self.post(
+                {
+                    "type": "handlers:add",
+                    "handlers": [self.app._handlers[target_id][0].to_json() for target_id in target_ids],
+                }
+            )
+
+            return result
+
+        async def _on_memory_read(self, payload: dict) -> None:
+            data = self.app._tracer.read_memory(payload["address"], payload["size"])
+            return list(data) if data is not None else None
+
+        async def _on_symbols_resolve_addresses(self, payload: dict) -> None:
+            names = self.app._tracer.resolve_addresses(payload["addresses"])
+            return {"names": names}
 
     app = TracerApplication()
     app.run()
@@ -278,6 +525,16 @@ class TracerProfileBuilder:
             self._spec.append(("exclude", "objc-method", f))
         return self
 
+    def include_swift_func(self, *function_name_globs: str) -> "TracerProfileBuilder":
+        for f in function_name_globs:
+            self._spec.append(("include", "swift-func", f))
+        return self
+
+    def exclude_swift_func(self, *function_name_globs: str) -> "TracerProfileBuilder":
+        for f in function_name_globs:
+            self._spec.append(("exclude", "swift-func", f))
+        return self
+
     def include_java_method(self, *function_name_globs: str) -> "TracerProfileBuilder":
         for f in function_name_globs:
             self._spec.append(("include", "java-method", f))
@@ -311,15 +568,17 @@ class Tracer:
         init_scripts=[],
         log_handler: Callable[[str, str], None] = None,
     ) -> None:
+        self.main_module = None
         self._reactor = reactor
         self._repository = repository
         self._profile = profile
         self._script: Optional[frida.core.Script] = None
+        self._schedule_on_message = None
         self._agent = None
         self._init_scripts = init_scripts
         self._log_handler = log_handler
 
-    def start_trace(self, session: frida.core.Session, stage, parameters, runtime, ui) -> None:
+    def start_trace(self, session: frida.core.Session, stage, parameters, runtime, ui: UI) -> None:
         def on_create(*args) -> None:
             ui.on_trace_handler_create(*args)
 
@@ -331,39 +590,57 @@ class Tracer:
         self._repository.on_load(on_load)
 
         def on_update(target, handler, source) -> None:
-            self._agent.update(target.identifier, target.display_name, handler)
+            self._agent.update_handler_code(target.identifier, target.display_name, handler)
 
         self._repository.on_update(on_update)
 
-        def on_message(message, data) -> None:
-            self._reactor.schedule(lambda: self._on_message(message, data, ui))
+        self._schedule_on_message = lambda message, data: self._reactor.schedule(
+            lambda: self._on_message(message, data, ui)
+        )
 
         ui.on_trace_progress("initializing")
         data_dir = os.path.dirname(__file__)
         with codecs.open(os.path.join(data_dir, "tracer_agent.js"), "r", "utf-8") as f:
             source = f.read()
-        runtime = "v8" if runtime == "v8" else "qjs"
         script = session.create_script(name="tracer", source=source, runtime=runtime)
 
         self._script = script
         script.set_log_handler(self._log_handler)
-        script.on("message", on_message)
+        script.on("message", self._schedule_on_message)
         ui.on_script_created(script)
         script.load()
 
-        self._agent = script.exports
+        self._agent = script.exports_sync
 
         raw_init_scripts = [{"filename": script.filename, "source": script.source} for script in self._init_scripts]
-        self._agent.init(stage, parameters, raw_init_scripts, self._profile.spec)
+        self.process = self._agent.init(stage, parameters, raw_init_scripts, self._profile.spec)
 
     def stop(self) -> None:
+        self._repository.close()
+
         if self._script is not None:
+            self._script.off("message", self._schedule_on_message)
             try:
                 self._script.unload()
             except:
                 pass
             self._script = None
 
+    def update_handler_config(self, identifier: int, config: dict) -> None:
+        return self._agent.update_handler_config(identifier, config)
+
+    def stage_targets(self, profile: TracerProfile) -> List:
+        return self._agent.stage_targets(profile.spec)
+
+    def commit_targets(self, identifier: Optional[int]) -> dict:
+        return self._agent.commit_targets(identifier)
+
+    def read_memory(self, address: str, size: int) -> bytes:
+        return self._agent.read_memory(address, size)
+
+    def resolve_addresses(self, addresses: List[str]) -> List[str]:
+        return self._agent.resolve_addresses(addresses)
+
     def _on_message(self, message, data, ui) -> None:
         handled = False
 
@@ -384,8 +661,8 @@ class Tracer:
     def _try_handle_message(self, mtype, params, data, ui) -> False:
         if mtype == "events:add":
             events = [
-                (timestamp, thread_id, depth, message)
-                for target_id, timestamp, thread_id, depth, message in params["events"]
+                (target_id, timestamp, thread_id, depth, caller, backtrace, message)
+                for target_id, timestamp, thread_id, depth, caller, backtrace, message in params["events"]
             ]
             ui.on_trace_events(events)
             return True
@@ -401,11 +678,20 @@ class Tracer:
             next_id = base_id
             for scope in params["scopes"]:
                 scope_name = scope["name"]
+                addresses = scope.get("addresses")
+                i = 0
                 for member_name in scope["members"]:
-                    target = TraceTarget(next_id, flavor, scope_name, member_name)
+                    if isinstance(member_name, list):
+                        name, display_name, address = member_name
+                    else:
+                        name = member_name
+                        display_name = member_name
+                    address = int(addresses[i], 16) if addresses is not None else None
+                    target = TraceTarget(next_id, flavor, scope_name, name, display_name, address)
                     next_id += 1
                     handler = repo.ensure_handler(target)
                     scripts.append(handler)
+                    i += 1
 
             self._script.post(response)
 
@@ -431,17 +717,23 @@ class Tracer:
         return False
 
 
+@dataclass
 class TraceTarget:
-    def __init__(self, identifier, flavor, scope, name: Union[str, List[str]]) -> None:
-        self.identifier = identifier
-        self.flavor = flavor
-        self.scope = scope
-        if isinstance(name, list):
-            self.name = name[0]
-            self.display_name = name[1]
-        else:
-            self.name = name
-            self.display_name = name
+    identifier: int
+    flavor: str
+    scope: str
+    name: str
+    display_name: str
+    address: Optional[int]
+
+    def to_json(self) -> dict:
+        return {
+            "id": self.identifier,
+            "flavor": self.flavor,
+            "scope": self.scope,
+            "display_name": self.display_name,
+            "address": hex(self.address) if self.address is not None else None,
+        }
 
     def __str__(self) -> str:
         return self.display_name
@@ -453,6 +745,12 @@ class Repository:
         self._on_load_callback: Optional[Callable[[TraceTarget, str, str], None]] = None
         self._on_update_callback: Optional[Callable[[TraceTarget, str, str], None]] = None
         self._decorate = False
+        self._manpages = None
+
+    def close(self) -> None:
+        self._on_create_callback = None
+        self._on_load_callback = None
+        self._on_update_callback = None
 
     def ensure_handler(self, target: TraceTarget):
         raise NotImplementedError("not implemented")
@@ -482,94 +780,34 @@ class Repository:
             self._on_update_callback(target, handler, source)
 
     def _create_stub_handler(self, target: TraceTarget, decorate: bool) -> str:
+        if target.flavor == "insn":
+            return self._create_stub_instruction_handler(target, decorate)
         if target.flavor == "java":
             return self._create_stub_java_handler(target, decorate)
-        else:
-            return self._create_stub_native_handler(target, decorate)
+        return self._create_stub_native_handler(target, decorate)
+
+    def _create_stub_instruction_handler(self, target: TraceTarget, decorate: bool) -> str:
+        return """\
+/*
+ * Auto-generated by Frida.
+ *
+ * For full API reference, see: https://frida.re/docs/javascript-api/
+ */
+
+defineHandler(function (log, args, state) {
+  log(`%(display_name)s hit! sp=${this.context.sp}`);
+});
+""" % {
+            "display_name": target.display_name
+        }
 
     def _create_stub_native_handler(self, target: TraceTarget, decorate: bool) -> str:
         if target.flavor == "objc":
-            state = {"index": 2}
-
-            def objc_arg(m):
-                index = state["index"]
-                r = ":${args[%d]} " % index
-                state["index"] = index + 1
-                return r
-
-            log_str = "`" + re.sub(r":", objc_arg, target.display_name) + "`"
-            if log_str.endswith("} ]`"):
-                log_str = log_str[:-3] + "]`"
+            log_str = self._create_objc_logging_code(target)
+        elif target.flavor == "swift":
+            log_str = self._create_swift_logging_code(target, decorate)
         else:
-            for man_section in (2, 3):
-                args = []
-                try:
-                    with open(os.devnull, "w") as devnull:
-                        man_argv = ["man"]
-                        if platform.system() != "Darwin":
-                            man_argv.extend(["-E", "UTF-8"])
-                        man_argv.extend(["-P", "col -b", str(man_section), target.name])
-                        output = subprocess.check_output(man_argv, stderr=devnull)
-                    match = re.search(
-                        r"^SYNOPSIS(?:.|\n)*?((?:^.+$\n)* {5}\w+[ \*\n]*"
-                        + target.name
-                        + r"\((?:.+\,\s*?$\n)*?(?:.+\;$\n))(?:.|\n)*^DESCRIPTION",
-                        output.decode("UTF-8", errors="replace"),
-                        re.MULTILINE,
-                    )
-                    if match:
-                        decl = match.group(1)
-
-                        for argm in re.finditer(r"[\(,]\s*(.+?)\s*\b(\w+)(?=[,\)])", decl):
-                            typ = argm.group(1)
-                            arg = argm.group(2)
-                            if arg == "void":
-                                continue
-                            if arg == "...":
-                                args.append('", ..." +')
-                                continue
-
-                            read_ops = ""
-                            annotate_pre = ""
-                            annotate_post = ""
-
-                            normalized_type = re.sub(r"\s+", "", typ)
-                            if normalized_type.endswith("*restrict"):
-                                normalized_type = normalized_type[:-8]
-                            if normalized_type in ("char*", "constchar*"):
-                                read_ops = ".readUtf8String()"
-                                annotate_pre = '"'
-                                annotate_post = '"'
-
-                            arg_index = len(args)
-
-                            args.append(
-                                "%(arg_name)s=%(annotate_pre)s${args[%(arg_index)s]%(read_ops)s}%(annotate_post)s"
-                                % {
-                                    "arg_name": arg,
-                                    "arg_index": arg_index,
-                                    "read_ops": read_ops,
-                                    "annotate_pre": annotate_pre,
-                                    "annotate_post": annotate_post,
-                                }
-                            )
-                        break
-                except Exception:
-                    pass
-
-            if decorate:
-                module_string = " [%s]" % os.path.basename(target.scope)
-            else:
-                module_string = ""
-
-            if len(args) == 0:
-                log_str = "'%(name)s()%(module_string)s'" % {"name": target.name, "module_string": module_string}
-            else:
-                log_str = "`%(name)s(%(args)s)%(module_string)s`" % {
-                    "name": target.name,
-                    "args": ", ".join(args),
-                    "module_string": module_string,
-                }
+            log_str = self._create_cstyle_logging_code(target, decorate)
 
         return """\
 /*
@@ -579,42 +817,59 @@ class Repository:
  * For full API reference, see: https://frida.re/docs/javascript-api/
  */
 
-{
-  /**
-   * Called synchronously when about to call %(display_name)s.
-   *
-   * @this {object} - Object allowing you to store state for use in onLeave.
-   * @param {function} log - Call this function with a string to be presented to the user.
-   * @param {array} args - Function arguments represented as an array of NativePointer objects.
-   * For example use args[0].readUtf8String() if the first argument is a pointer to a C string encoded as UTF-8.
-   * It is also possible to modify arguments by assigning a NativePointer object to an element of this array.
-   * @param {object} state - Object allowing you to keep state across function calls.
-   * Only one JavaScript function will execute at a time, so do not worry about race-conditions.
-   * However, do not use this to store function arguments across onEnter/onLeave, but instead
-   * use "this" which is an object for keeping state local to an invocation.
-   */
+defineHandler({
   onEnter(log, args, state) {
     log(%(log_str)s);
   },
 
-  /**
-   * Called synchronously when about to return from %(display_name)s.
-   *
-   * See onEnter for details.
-   *
-   * @this {object} - Object allowing you to access state stored in onEnter.
-   * @param {function} log - Call this function with a string to be presented to the user.
-   * @param {NativePointer} retval - Return value represented as a NativePointer object.
-   * @param {object} state - Object allowing you to keep state across function calls.
-   */
   onLeave(log, retval, state) {
   }
-}
+});
 """ % {
             "display_name": target.display_name,
             "log_str": log_str,
         }
 
+    def _create_cstyle_logging_code(self, target: TraceTarget, decorate: bool) -> str:
+        if decorate:
+            module_string = " [%s]" % os.path.basename(target.scope)
+        else:
+            module_string = ""
+
+        args = self._generate_cstyle_argument_logging_code(target)
+        if len(args) == 0:
+            code = "'%(name)s()%(module_string)s'" % {"name": target.name, "module_string": module_string}
+        else:
+            code = "`%(name)s(%(args)s)%(module_string)s`" % {
+                "name": target.name,
+                "args": ", ".join(args),
+                "module_string": module_string,
+            }
+
+        return code
+
+    def _create_objc_logging_code(self, target: TraceTarget) -> str:
+        state = {"index": 2}
+
+        def objc_arg(m):
+            index = state["index"]
+            r = ":${args[%d]} " % index
+            state["index"] = index + 1
+            return r
+
+        code = "`" + re.sub(r":", objc_arg, target.display_name) + "`"
+        if code.endswith("} ]`"):
+            code = code[:-3] + "]`"
+
+        return code
+
+    def _create_swift_logging_code(self, target: TraceTarget, decorate: bool) -> str:
+        if decorate:
+            module_string = " [%s]" % os.path.basename(target.scope)
+        else:
+            module_string = ""
+        return "'%(name)s()%(module_string)s'" % {"name": target.name, "module_string": module_string}
+
     def _create_stub_java_handler(self, target: TraceTarget, decorate) -> str:
         return """\
 /*
@@ -623,7 +878,7 @@ class Repository:
  * For full API reference, see: https://frida.re/docs/javascript-api/
  */
 
-{
+defineHandler({
   /**
    * Called synchronously when about to call %(display_name)s.
    *
@@ -651,11 +906,167 @@ class Repository:
       log(`<= ${JSON.stringify(retval)}`);
     }
   }
-}
+});
 """ % {
             "display_name": target.display_name
         }
 
+    def _generate_cstyle_argument_logging_code(self, target: TraceTarget) -> List[str]:
+        if self._manpages is None:
+            self._manpages = {}
+            try:
+                manroots = [
+                    Path(d)
+                    for d in subprocess.run(["manpath"], stdout=subprocess.PIPE, encoding="utf-8", check=True)
+                    .stdout.strip()
+                    .split(":")
+                ]
+                for section in (2, 3):
+                    for manroot in manroots:
+                        mandir = manroot / f"man{section}"
+                        if not mandir.exists():
+                            continue
+                        raw_section = str(section)
+                        for entry in mandir.iterdir():
+                            tokens = entry.name.split(".")
+                            if len(tokens) < 2:
+                                continue
+                            if not tokens[1].startswith(raw_section):
+                                continue
+                            name = tokens[0]
+                            if name in self._manpages:
+                                continue
+                            self._manpages[name] = (entry, section)
+            except:
+                return []
+
+        man_entry = self._manpages.get(target.name)
+        if man_entry is None:
+            return []
+        man_location, man_section = man_entry
+
+        try:
+            args = []
+            cfunc = next(f for f in self._read_manpage(man_location) if f.name == target.name)
+            for arg in cfunc.arguments:
+                if arg == "void":
+                    continue
+                if arg.startswith("..."):
+                    args.append("...")
+                    continue
+
+                tokens = arg.split(" ")
+
+                arg_type = "".join(tokens[:-1])
+
+                arg_name = tokens[-1]
+                if arg_name.startswith("*"):
+                    arg_type += "*"
+                    arg_name = arg_name[1:]
+                elif arg_name.endswith("]"):
+                    arg_type += "*"
+                    arg_name = arg_name[: arg_name.index("[")]
+
+                read_ops = ""
+                annotate_pre = ""
+                annotate_post = ""
+
+                if arg_type.endswith("*restrict"):
+                    arg_type = arg_type[:-8]
+                if arg_type in ("char*", "constchar*"):
+                    read_ops = ".readUtf8String()"
+                    annotate_pre = '"'
+                    annotate_post = '"'
+
+                arg_index = len(args)
+
+                args.append(
+                    "%(arg_name)s=%(annotate_pre)s${args[%(arg_index)s]%(read_ops)s}%(annotate_post)s"
+                    % {
+                        "arg_name": arg_name,
+                        "arg_index": arg_index,
+                        "read_ops": read_ops,
+                        "annotate_pre": annotate_pre,
+                        "annotate_post": annotate_post,
+                    }
+                )
+            return args
+        except Exception:
+            return []
+
+    def _read_manpage(self, man_location: Path) -> Generator[CFuncSpec]:
+        if man_location.suffix == ".gz":
+            man_file = gzip.open(man_location, "rt", encoding="utf-8", errors="replace")
+        else:
+            man_file = open(man_location, "r", encoding="utf-8", errors="replace")
+        with man_file:
+            man_data = man_file.read()
+
+        manpage_format = "gnu"
+        synopsis_lines = []
+        found_synopsis = False
+        in_multiline = False
+        for raw_line in man_data.split("\n"):
+            line = raw_line.strip()
+            if line.startswith(".so "):
+                redirected_location = man_location.parent.parent / Path(line[4:])
+                if not redirected_location.exists():
+                    redirected_location = redirected_location.parent / (redirected_location.name + ".gz")
+                yield from self._read_manpage(redirected_location)
+                return
+            if not found_synopsis and "SYNOPSIS" in line:
+                found_synopsis = True
+                continue
+            elif found_synopsis and line.endswith("DESCRIPTION"):
+                break
+            elif not found_synopsis:
+                continue
+            if line.startswith(".Fn ") or line.startswith(".Fo "):
+                manpage_format = "bsd"
+            escaped_newline = line.endswith("\\")
+            if escaped_newline:
+                line = line[:-1]
+            if in_multiline:
+                synopsis_lines[-1] += line
+            else:
+                synopsis_lines.append(line)
+            in_multiline = escaped_newline
+
+        if manpage_format == "gnu":
+            raw_synopsis = "\n".join(synopsis_lines)
+            synopsis = (
+                MANPAGE_CONTROL_CHARS.sub("", raw_synopsis).replace("\n", " ").replace(" [", "[").replace(" ]", "]")
+            )
+
+            for m in MANPAGE_FUNCTION_PROTOTYPE.finditer(synopsis):
+                name = m.group(1)
+                signature = m.group(2)
+                args = [a.strip() for a in signature.split(",")]
+                yield CFuncSpec(name, args)
+        else:
+            name = None
+            args = None
+            for line in synopsis_lines:
+                tokens = line.split(" ", maxsplit=1)
+                directive = tokens[0]
+                data = tokens[1] if len(tokens) == 2 else None
+                if directive == ".Fn":
+                    argv = shlex.split(data)
+                    yield CFuncSpec(argv[0], argv[1:])
+                elif directive == ".Fo":
+                    name = data
+                    args = []
+                elif directive == ".Fa":
+                    args.append(shlex.split(data)[0])
+                elif directive == ".Fc":
+                    yield CFuncSpec(name, args)
+
+
+@dataclass
+class CFuncSpec:
+    name: str
+    arguments: List[str]
+
 
 class MemoryRepository(Repository):
     def __init__(self) -> None:
@@ -685,6 +1096,16 @@ class FileRepository(Repository):
         self._repo_monitors = {}
         self._decorate = decorate
 
+    def close(self) -> None:
+        for monitor in self._repo_monitors.values():
+            try:
+                monitor.disable()
+            except:
+                pass
+        self._repo_monitors.clear()
+
+        super().close()
+
     def ensure_handler(self, target: TraceTarget) -> str:
         entry = self._handler_by_id.get(target.identifier)
         if entry is not None:
@@ -723,6 +1144,15 @@ class FileRepository(Repository):
 
         return handler
 
+    def update_handler(self, target: TraceTarget, handler: str) -> None:
+        _, _, handler_file = self._handler_by_id.get(target.identifier)
+        entry = (target, handler, handler_file)
+        self._handler_by_id[target.identifier] = entry
+        self._handler_by_file[handler_file] = entry
+        self._notify_update(target, handler, handler_file)
+
+        Path(handler_file).write_text(handler, encoding="utf-8")
+
     def _ensure_monitor(self, handler_file) -> None:
         handler_dir = os.path.dirname(handler_file)
         monitor = self._repo_monitors.get(handler_dir)
